package edu.wlu.cs.levy.RAAM;

import java.util.Vector;
import java.util.HashSet;
import java.util.Iterator;

 /**
  * Tree is a class supporting trees of arbitrary arity.
  *
  * @author      Matthew Grieco
  * @author      Simon Levy
  * @version     %I%, %G%
  * @since JDK1.2 
  */

public class Tree {
   	
    public String value;
    public Tree [] branches;
   

    /**
     * Creates a tree by parsing a parenthesized expression. If expression is
     * not a well-formed tree, behavior is undefined.
     *
     * @param s the parenthesized expression
     *
     */
    public Tree(String s) {

        // terminal
	if (s.indexOf('(') < 0) {
	    value = new String(s);
	}

	// non-terminal
	else {

          // initial conditions: empty stack, start after first paren,
          // no subtrees
          int stack = 0;
          int last = 1;
          int i = 1;
	  Vector v = new Vector();

          // loop over characters between parens
          while (i < s.length()) {

            // current character
            char c = s.charAt(i);
     
            // use stack counter to track sub-trees
            if (c == '(')
	        stack++;
            else if (c == ')')
	        stack--;

            // on empty stack, build a branch by recurring on subtree
            if (stack == 0) {

	      // skip to end of current token
	      while (s.charAt(i) != ' ' && 
                     s.charAt(i) != '(' &&
                     s.charAt(i) != ')')
	          i++;

	      // keep paren at end if found at beginning
	      int curr = s.charAt(last)=='(' ? i+1 : i;

	      // not sure why this constraint gets violated
	      if (curr > last) {

	        // recur on token
	        String ssub = s.substring(last, curr);
	        Tree child = new Tree(ssub);

	        // add child to list
	        v.add(child);
	      }

	      // track last ending position
              last = i + 1;
	    }    

	    i++;
	  }

          value = null;
          branches = new Tree[v.size()];
	  for (int j=0; j<v.size(); ++j) {
	      branches[j] = (Tree)v.elementAt(j);
	  }

	} // else (non-term)


    }


    /**
     * Creates tree from a shallow copy of multiple branches. 
     *
     * @param b the branches
     *
     */
    public Tree(Tree [] b) {
	
	value = null;
	branches = b;
    }


    /**
     * Creates a tree from two branches.
     *
     * @param l the left branch
     * @param r the right branch
     *
     */
    public Tree(Tree l, Tree r) {

	this(lr_array(l,r));
    }

    /**
     * Determines whether tree is terminal (has non-empty value).
     *
     * @return true if tree is terminal, false otherwise
     *
     */
    public boolean isTerminal() {
        return (value != null);
    }

    /**
     * Overrides Object.toString() method.
     *
     * @return a parenthesized expression representing the tree
     *
     */
    public String toString () {

        if (isTerminal()) {
	    return value.toString();
	}
	else {
            String s = "";
            for (int i = 0; i<branches.length; i++){
                s = s + branches[i].toString();
		if (i<branches.length-1) s += "."; // for Matlab
            }
            s = s.trim();
            return "(" + s + ")";
	}
    }


    /**
     * Overrides Object.equals() method.
     *
     */
    public boolean equals(Object o) {

	return !getClass().equals(o.getClass()) ? 
	    false :
	    toString().equals(o.toString());
    }

    /**
     * Overrides Object.hashCode() method.
     *
     */
    public int hashCode() {

	return toString().hashCode();
    }


    /**
     * Returns the string in the value field of the tree.
     *
     * @return the string in the value field of the tree
     *
     */
    public String getValue () {
	return value;
    }

    /**
     * Returns a branch of the tree.
     * @param n branch index (first = 0)
     *
     * @return the <i>n</i>th branch of the tree
     *
     */
    public Tree getBranch (int n) {
	return  branches[n];
    }


    /**
     * Returns the arity (number of branches) of the tree.
     *
     * @return the arity of the tree
     *
     */
    public int arity () {
	return  branches.length;
    }

    /**
     * Returns the depth of the tree.
     *
     * @return the depth of the tree
     *
     */
    public int depth () {
	if (isTerminal()) {
	    return 0;
	}
	else {
	    int dmax = 0;
	    for (int i=0; i<branches.length; ++i) {
		int d = 1 + branches[i].depth();
		if (d > dmax) {
		    dmax = d;
		}
	    }
	    return dmax;
	}
    }

    /**
     * Returns the frontier of the tree.
     *
     * @return the frontier of the tree
     *
     */
    public String frontier () {
	if (isTerminal()) {
	    return value.toString();
	}
	else {
	    String s = "";
	    for (int i=0; i<branches.length; ++i) {
		s += branches[i].frontier();
		if (i < branches.length) {
		    s += " ";
		}
	    }
	    return s;
	}
    }

    /**
     * Returns all subtrees of the tree.
     *
     * @return the subtrees
     *
     */
    public Tree [] subtrees () {
	HashSet h = new HashSet();
	_subtrees(h);
	Tree [] subs = new Tree [h.size()];
	Iterator i = h.iterator();
	int j = 0;
	while (i.hasNext()) {
	    subs[j++] = (Tree)i.next();
	}
	return subs;
    }

    // put subtrees into a set
    private void _subtrees(HashSet h) {

	h.add(this);

	if (!isTerminal()) {
	    for (int i=0; i<branches.length; ++i) {
		branches[i]._subtrees(h);
	    }
	}
    }

    // puts left, right trees into array for constructor
    private static Tree [] lr_array(Tree l, Tree r) {
	Tree [] a = {l, r};
	return a;
    }

}

