package edu.wlu.cs.levy.RAAM;

import java.util.Hashtable;
import java.util.HashSet;
import java.util.Enumeration;
import java.util.Iterator;

import edu.wlu.cs.levy.SNARLI.SNVector;

 /**
  * Vectionary is a class supporting a mapping from strings to vectors
  * (double-precision arrays).
  *
  * @author      Simon Levy
  * @version     %I%, %G%
  * @since JDK1.2 
  */

public class Vectionary {

    private int size;         // fixed vector size
    private Hashtable table;  // string->vectors
    private Hashtable rtable; // vector->string

    /**
     * Creates a dictionary with vector entries of specified size.
     *
     * @param size vector size
     *
     */
    public Vectionary(int size) {

	this.size = size;
	table = new Hashtable();
	rtable = new Hashtable();
    }


    /**
     * Enters a string<->vector mapping into the dictionary
     *
     * @param s the string
     * @param p the vector
     * @throws IllegalArgumentException if array length not equal to
     * length specified in call to constructor
     *
     */
    public void enter(String s, double [] p) 

	throws IllegalArgumentException {

	if (p.length != size) {
	    throw new IllegalArgumentException("Array length mismatch");
	}
	
	HashSet hs = table.containsKey(s) ? 
	    (HashSet)table.get(s) :
	    new HashSet();

	hs.add(new SNVector(p));

	table.put(s, hs);

	rtable.put(new SNVector(p), s);
    }


    /**
     * Enters vectors->string mappings into the dictionary.  This is useful
     * when multiple vectors share the same string.
     *
     * @param s the string
     * @param pts the vectors
     * @throws IllegalArgumentException if array length not equal to
     * length specified in call to constructor
     *
     */
    public void enter(String s, double [][] pts) {
	for (int i=0; i<pts.length; ++i) {
	    enter(s, pts[i]);
	}
    }


    /**
     * Overrides Object.toString() method.
     *
     * @return a string representing the table for this dictionary
     *
     */
    public String toString() {

	String s = "";
	for (Enumeration e=table.keys(); e.hasMoreElements(); ) {
	    Object o = e.nextElement();
	    s = s + o + " : \n";
	    HashSet hs = (HashSet)table.get(o);
	    for (Iterator i=hs.iterator(); i.hasNext(); ) {
		s = s + "    " + i.next() + "\n";
	    }
	}
	return s;
    }
 

    /**
     * Returns vector(s) keyed by string.
     *
     * @param s the string
     * @return array of vectors keyed by string
     * @throws IllegalArgumentException if there is no such key
     *
     */
    public double [][] lookup(String s) 

	throws IllegalArgumentException {

	if (!table.containsKey(s)) {
	    throw new IllegalArgumentException("No vectors matching key '" + 
					       s + "'");
	}

	HashSet hs = (HashSet)table.get(s);
	double [][] pts = new double [hs.size()][];
	int j = 0;
	for (Iterator i=hs.iterator(); i.hasNext(); ) {
	    SNVector v = (SNVector)i.next();
	    pts[j++] = v.getValues();
	}
	return pts;
    }

    /**
     * Returns string that keys vector.
     *
     * @param p the vector
     * @return string that is the key for that vector
     * @throws IllegalArgumentException if there is no key for that vector
     *
     */
    public String lookup(double [] p) 

	throws IllegalArgumentException {

	SNVector fp = new SNVector(p);

	if (!rtable.containsKey(fp)) {
	    throw new IllegalArgumentException("No key matching vector " + fp);
	}
	
	return (String)rtable.get(fp);
    }


    /**
     * Returns vector size.
     *
     * @return vector size
     *
     */
    public int width() {
	
	return size;
    }

    /**
     * Returns number of entries.
     *
     * @return number of entries
     *
     */
    public int count() {
	
	return table.size();
    }

    // Return entry having key closest to argument.  Must be less than 
    // EPSILON Euclidean distance.
    protected String closest(double [] code, double epsilon) {
	String s = null;
	if (table.size() > 0) {
	    SNVector cv = new SNVector(code);
	    double mindist = Double.POSITIVE_INFINITY;
	    for (Enumeration e=rtable.keys(); e.hasMoreElements(); ) {
		SNVector key = (SNVector)e.nextElement();
		double dist = cv.eucDist(key);
		if (dist < mindist && dist < epsilon) {
		    mindist = dist;
		    s = (String)rtable.get(key);
		}
	    }
	}
	return s;
    }
}

