package edu.wlu.cs.levy.SNARLI;

import java.util.Random;
import java.awt.Point;

 /**
  * SOM implements Kohonen's Self-Organizing Map.
  *
  * @author      Simon Levy
  * @version     %I%, %G%
  * @since JDK1.2 
  */
public class SOM extends Map2D {

    // initial, final width parameters
    private double _sigma_i, _sigma_f;

    /** 
     * Builds SOM with random initial weights.
     * @param X data to learn
     * @param N1 number of rows
     * @param N2 number of columns
     * @param P probability density function over X
     * @param rand random-number generator
     * @throws IllegalArgumentException if X has fewer than two dimensions
     * @throws IllegalArgumentException if X and P have different lengths
     * @throws IllegalArgumentException if elements of P do not sum to 1
     */
    public SOM(double [][] X, int N1, int N2, double [] P, Random rand) 
        throws IllegalArgumentException {
	super(X, N1, N2, P, rand);
    }

    /** 
     * Builds SOM with random initial weights.
     * @param X data to learn
     * @param P probability density function over X
     * @param N1 number of rows
     * @param N2 number of columns
     * @throws IllegalArgumentException if x and P have different lengths
     * @throws IllegalArgumentException if elements of P do not sum to 1
     */
    public SOM(double [][] X, int N1, int N2, double [] P) 
	throws IllegalArgumentException {
        this(X, N1, N2, P, new Random());
    }


    /** 
     * Builds SOM with random initial weights.
     * @param X data to learn
     * @param P probability density function over X
     * @param N1 number of rows
     * @param N2 number of columns
     * @param seed seed for random-number generator
     * @throws IllegalArgumentException if X and P have different lengths
     * @throws IllegalArgumentException if elements of P do not sum to 1
     */
    public SOM(double [][] X, int N1, int N2, double [] P , long seed)  
        throws IllegalArgumentException {
        this(X, N1, N2, P, new Random(seed));
    }


    /** 
     * Builds SOM with random initial weights and uniform probability
     * density function.
     * @param X data to learn
     * @param N1 number of rows
     * @param N2 number of columns
     */
    public SOM(double [][] X, int N1, int N2) {
        this(X, N1, N2, new Random());
    }


    /** 
     * Builds SOM with random initial weights and uniform probability
     * density function.
     * @param X data to learn
     * @param N1 number of rows
     * @param N2 number of columns
     * @param seed seed for random-number generator
     */
    public SOM(double [][] X, int N1, int N2, long seed) {
        this(X, N1, N2, new Random(seed));
    }

    /** 
     * Builds SOM with random initial weights and uniform probability
     * density function.
     * @param X data to learn
     * @param N1 number of rows
     * @param N2 number of columns
     * @param rand random-number generator
     */
    public SOM(double [][] X, int N1, int N2, Random rand) {
        this(X, N1, N2, updf(X), rand);
    }


    /** 
     * Runs learning iterations. Learning rate e(t) = ei(ef/ei)^{t/tmax}.
     * @param e_i initial learning rate
     * @param e_f final learning rate
     * @param sigma_i initial width parameter
     * @param sigma_f final width parameter
     * @param tmax total number of time-steps to run
     */
    public void learn(double ei, double ef, double sigma_i, double sigma_f,
		      int tmax) {

	_sigma_i = sigma_i;
	_sigma_f = sigma_f;

	super.learn(ei, ef, tmax, "Learning");
    }

    // Initialize A to contain N = _N1*_N2 units Ci, with reference
    // vectors Wci in R^n chosen randomly according to P(X).
    protected void initA() {
	for (int i=0; i<_N1; ++i) {
	    for (int j=0; j<_N2; ++j) {
		_A[i][j] = randomSignal();
	    }
	}
    }

    // returns width parameter
    protected double sigma(int t, int tmax) {
	return scale(_sigma_i, _sigma_f, t, tmax);
    }

}

